@@ -12,25 +12,51 @@ struct ReinterpretArray{T,N,S,A<:AbstractArray{S},IsReshaped} <: AbstractArray{T
     writable::Bool
 
     function throwbits(S::Type, T::Type, U::Type)
-        @_noinline_meta
+        @noinline
         throw(ArgumentError("cannot reinterpret `$(S)` as `$(T)`, type `$(U)` is not a bits type"))
     end
     function throwsize0(S::Type, T::Type, msg)
-        @_noinline_meta
+        @noinline
         throw(ArgumentError("cannot reinterpret a zero-dimensional `$(S)` array to `$(T)` which is of a $msg size"))
     end
+    function throwsingleton(S::Type, T::Type)
+        @noinline
+        throw(ArgumentError("cannot reinterpret a `$(S)` array to `$(T)` which is a singleton type"))
+    end
 
     global reinterpret
+
+    """
+        reinterpret(T::DataType, A::AbstractArray)
+
+    Construct a view of the array with the same binary data as the given
+    array, but with `T` as element type.
+
+    This function also works on "lazy" array whose elements are not computed until they are explicitly retrieved.
+    For instance, `reinterpret` on the range `1:6` works similarly as on the dense vector `collect(1:6)`:
+
+    ```jldoctest
+    julia> reinterpret(Float32, UInt32[1 2 3 4 5])
+    1×5 reinterpret(Float32, ::Matrix{UInt32}):
+    1.0f-45  3.0f-45  4.0f-45  6.0f-45  7.0f-45
+
+    julia> reinterpret(Complex{Int}, 1:6)
+    3-element reinterpret(Complex{$Int}, ::UnitRange{$Int}):
+    1 + 2im
+    3 + 4im
+    5 + 6im
+    ```
+    """
     function reinterpret(::Type{T}, a::A) where {T,N,S,A<:AbstractArray{S, N}}
         function thrownonint(S::Type, T::Type, dim)
-            @_noinline_meta
+            @noinline
             throw(ArgumentError("""
                 cannot reinterpret an `$(S)` array to `$(T)` whose first dimension has size `$(dim)`.
                 The resulting array would have non-integral first dimension.
                 """))
         end
         function throwaxes1(S::Type, T::Type, ax1)
-            @_noinline_meta
+            @noinline
             throw(ArgumentError("cannot reinterpret a `$(S)` array to `$(T)` when the first axis is $ax1. Try reshaping first."))
         end
         isbitstype(T) || throwbits(S, T, T)
@@ -39,7 +65,11 @@ struct ReinterpretArray{T,N,S,A<:AbstractArray{S},IsReshaped} <: AbstractArray{T
         if N != 0 && sizeof(S) != sizeof(T)
             ax1 = axes(a)[1]
             dim = length(ax1)
-            rem(dim*sizeof(S),sizeof(T)) == 0 || thrownonint(S, T, dim)
+            if issingletontype(T)
+                issingletontype(S) || throwsingleton(S, T)
+            else
+                rem(dim*sizeof(S),sizeof(T)) == 0 || thrownonint(S, T, dim)
+            end
             first(ax1) == 1 || throwaxes1(S, T, ax1)
         end
         readable = array_subpadding(T, S)
@@ -51,25 +81,31 @@ struct ReinterpretArray{T,N,S,A<:AbstractArray{S},IsReshaped} <: AbstractArray{T
     # With reshaping
     function reinterpret(::typeof(reshape), ::Type{T}, a::A) where {T,S,A<:AbstractArray{S}}
         function throwintmult(S::Type, T::Type)
-            @_noinline_meta
+            @noinline
             throw(ArgumentError("`reinterpret(reshape, T, a)` requires that one of `sizeof(T)` (got $(sizeof(T))) and `sizeof(eltype(a))` (got $(sizeof(S))) be an integer multiple of the other"))
         end
         function throwsize1(a::AbstractArray, T::Type)
-            @_noinline_meta
+            @noinline
             throw(ArgumentError("`reinterpret(reshape, $T, a)` where `eltype(a)` is $(eltype(a)) requires that `axes(a, 1)` (got $(axes(a, 1))) be equal to 1:$(sizeof(T) ÷ sizeof(eltype(a))) (from the ratio of element sizes)"))
         end
+        function throwfromsingleton(S, T)
+            @noinline
+            throw(ArgumentError("`reinterpret(reshape, $T, a)` where `eltype(a)` is $S requires that $T be a singleton type, since $S is one"))
+        end
         isbitstype(T) || throwbits(S, T, T)
         isbitstype(S) || throwbits(S, T, S)
         if sizeof(S) == sizeof(T)
             N = ndims(a)
         elseif sizeof(S) > sizeof(T)
+            issingletontype(T) && throwsingleton(S, T)
             rem(sizeof(S), sizeof(T)) == 0 || throwintmult(S, T)
             N = ndims(a) + 1
         else
+            issingletontype(S) && throwfromsingleton(S, T)
             rem(sizeof(T), sizeof(S)) == 0 || throwintmult(S, T)
             N = ndims(a) - 1
             N > -1 || throwsize0(S, T, "larger")
-            axes(a, 1) == Base.OneTo(sizeof(T) ÷ sizeof(S)) || throwsize1(a, T)
+            axes(a, 1) == OneTo(sizeof(T) ÷ sizeof(S)) || throwsize1(a, T)
         end
         readable = array_subpadding(T, S)
         writable = array_subpadding(S, T)
@@ -134,33 +170,35 @@ StridedVector{T} = StridedArray{T,1}
 StridedMatrix{T} = StridedArray{T,2}
 StridedVecOrMat{T} = Union{StridedVector{T}, StridedMatrix{T}}
 
-# the definition of strides for Array{T,N} is tuple() if N = 0, otherwise it is
-# a tuple containing 1 and a cumulative product of the first N-1 sizes
-# this definition is also used for StridedReshapedArray and StridedReinterpretedArray
-# which have the same memory storage as Array
-stride(a::Union{DenseArray,StridedReshapedArray,StridedReinterpretArray}, i::Int) = _stride(a, i)
-
-function stride(a::ReinterpretArray, i::Int)
-    a.parent isa StridedArray || ArgumentError("Parent must be strided.") |> throw
-    return _stride(a, i)
-end
-
-function _stride(a, i)
-    if i > ndims(a)
-        return length(a)
-    end
-    s = 1
-    for n = 1:(i-1)
-        s *= size(a, n)
-    end
-    return s
-end
-
-function strides(a::ReinterpretArray)
-    a.parent isa StridedArray || ArgumentError("Parent must be strided.") |> throw
-    size_to_strides(1, size(a)...)
-end
 strides(a::Union{DenseArray,StridedReshapedArray,StridedReinterpretArray}) = size_to_strides(1, size(a)...)
+stride(A::Union{DenseArray,StridedReshapedArray,StridedReinterpretArray}, k::Integer) =
+    k ≤ ndims(A) ? strides(A)[k] : length(A)
+
+function strides(a::ReinterpretArray{T,<:Any,S,<:AbstractArray{S},IsReshaped}) where {T,S,IsReshaped}
+    _checkcontiguous(Bool, a) && return size_to_strides(1, size(a)...)
+    stp = strides(parent(a))
+    els, elp = sizeof(T), sizeof(S)
+    els == elp && return stp # 0dim parent is also handled here.
+    IsReshaped && els < elp && return (1, _checked_strides(stp, els, elp)...)
+    stp[1] == 1 || throw(ArgumentError("Parent must be contiguous in the 1st dimension!"))
+    st′ = _checked_strides(tail(stp), els, elp)
+    return IsReshaped ? st′ : (1, st′...)
+end
+
+@inline function _checked_strides(stp::Tuple, els::Integer, elp::Integer)
+    if elp > els && rem(elp, els) == 0
+        N = div(elp, els)
+        return map(i -> N * i, stp)
+    end
+    drs = map(i -> divrem(elp * i, els), stp)
+    all(i->iszero(i[2]), drs) ||
+        throw(ArgumentError("Parent's strides could not be exactly divided!"))
+    map(first, drs)
+end
+
+_checkcontiguous(::Type{Bool}, A::ReinterpretArray) = _checkcontiguous(Bool, parent(A))
+
+similar(a::ReinterpretArray, T::Type, d::Dims) = similar(a.parent, T, d)
 
 function check_readable(a::ReinterpretArray{T, N, S} where N) where {T,S}
     # See comment in check_writable
@@ -211,12 +249,12 @@ SCartesianIndices2{K}(indices2::AbstractUnitRange{Int}) where {K} = (@assert K::
 eachindex(::IndexSCartesian2{K}, A::ReshapedReinterpretArray) where {K} = SCartesianIndices2{K}(eachindex(IndexLinear(), parent(A)))
 @inline function eachindex(style::IndexSCartesian2{K}, A::AbstractArray, B::AbstractArray...) where {K}
     iter = eachindex(style, A)
-    Base._all_match_first(C->eachindex(style, C), iter, B...) || Base.throw_eachindex_mismatch_indices(IndexSCartesian2{K}(), axes(A), axes.(B)...)
+    _all_match_first(C->eachindex(style, C), iter, B...) || throw_eachindex_mismatch_indices(IndexSCartesian2{K}(), axes(A), axes.(B)...)
     return iter
 end
 
 size(iter::SCartesianIndices2{K}) where K = (K, length(iter.indices2))
-axes(iter::SCartesianIndices2{K}) where K = (Base.OneTo(K), iter.indices2)
+axes(iter::SCartesianIndices2{K}) where K = (OneTo(K), iter.indices2)
 
 first(iter::SCartesianIndices2{K}) where {K} = SCartesianIndex2{K}(1, first(iter.indices2))
 last(iter::SCartesianIndices2{K}) where {K}  = SCartesianIndex2{K}(K, last(iter.indices2))
@@ -284,13 +322,13 @@ unaliascopy(a::ReshapedReinterpretArray{T}) where {T} = reinterpret(reshape, T,
 
 function size(a::NonReshapedReinterpretArray{T,N,S} where {N}) where {T,S}
     psize = size(a.parent)
-    size1 = div(psize[1]*sizeof(S), sizeof(T))
+    size1 = issingletontype(T) ? psize[1] : div(psize[1]*sizeof(S), sizeof(T))
     tuple(size1, tail(psize)...)
 end
 function size(a::ReshapedReinterpretArray{T,N,S} where {N}) where {T,S}
     psize = size(a.parent)
     sizeof(S) > sizeof(T) && return (div(sizeof(S), sizeof(T)), psize...)
-    sizeof(S) < sizeof(T) && return Base.tail(psize)
+    sizeof(S) < sizeof(T) && return tail(psize)
     return psize
 end
 size(a::NonReshapedReinterpretArray{T,0}) where {T} = ()
@@ -298,22 +336,31 @@ size(a::NonReshapedReinterpretArray{T,0}) where {T} = ()
 function axes(a::NonReshapedReinterpretArray{T,N,S} where {N}) where {T,S}
     paxs = axes(a.parent)
     f, l = first(paxs[1]), length(paxs[1])
-    size1 = div(l*sizeof(S), sizeof(T))
+    size1 = issingletontype(T) ? l : div(l*sizeof(S), sizeof(T))
     tuple(oftype(paxs[1], f:f+size1-1), tail(paxs)...)
 end
 function axes(a::ReshapedReinterpretArray{T,N,S} where {N}) where {T,S}
     paxs = axes(a.parent)
-    sizeof(S) > sizeof(T) && return (Base.OneTo(div(sizeof(S), sizeof(T))), paxs...)
-    sizeof(S) < sizeof(T) && return Base.tail(paxs)
+    sizeof(S) > sizeof(T) && return (OneTo(div(sizeof(S), sizeof(T))), paxs...)
+    sizeof(S) < sizeof(T) && return tail(paxs)
     return paxs
 end
 axes(a::NonReshapedReinterpretArray{T,0}) where {T} = ()
 
+has_offset_axes(a::ReinterpretArray) = has_offset_axes(a.parent)
+
 elsize(::Type{<:ReinterpretArray{T}}) where {T} = sizeof(T)
 unsafe_convert(::Type{Ptr{T}}, a::ReinterpretArray{T,N,S} where N) where {T,S} = Ptr{T}(unsafe_convert(Ptr{S},a.parent))
 
-@inline @propagate_inbounds getindex(a::NonReshapedReinterpretArray{T,0}) where {T} = reinterpret(T, a.parent[])
-@inline @propagate_inbounds getindex(a::ReinterpretArray) = a[1]
+@inline @propagate_inbounds function getindex(a::NonReshapedReinterpretArray{T,0,S}) where {T,S}
+    if isprimitivetype(T) && isprimitivetype(S)
+        reinterpret(T, a.parent[])
+    else
+        a[firstindex(a)]
+    end
+end
+
+@inline @propagate_inbounds getindex(a::ReinterpretArray) = a[firstindex(a)]
 
 @inline @propagate_inbounds function getindex(a::ReinterpretArray{T,N,S}, inds::Vararg{Int, N}) where {T,N,S}
     check_readable(a)
@@ -333,15 +380,11 @@ end
 
 @inline @propagate_inbounds function getindex(a::ReshapedReinterpretArray{T,N,S}, ind::SCartesianIndex2) where {T,N,S}
     check_readable(a)
-    n = sizeof(S) ÷ sizeof(T)
-    t = Ref{NTuple{n,T}}()
     s = Ref{S}(a.parent[ind.j])
-    GC.@preserve t s begin
-        tptr = Ptr{UInt8}(unsafe_convert(Ref{T}, t))
-        sptr = Ptr{UInt8}(unsafe_convert(Ref{S}, s))
-        _memcpy!(tptr, sptr, sizeof(S))
+    GC.@preserve s begin
+        tptr = Ptr{T}(unsafe_convert(Ref{S}, s))
+        return unsafe_load(tptr, ind.i)
     end
-    return t[][ind.i]
 end
 
 @inline _memcpy!(dst, src, n) = ccall(:memcpy, Cvoid, (Ptr{UInt8}, Ptr{UInt8}, Csize_t), dst, src, n)
@@ -349,33 +392,45 @@ end
 @inline @propagate_inbounds function _getindex_ra(a::NonReshapedReinterpretArray{T,N,S}, i1::Int, tailinds::TT) where {T,N,S,TT}
     # Make sure to match the scalar reinterpret if that is applicable
     if sizeof(T) == sizeof(S) && (fieldcount(T) + fieldcount(S)) == 0
+        if issingletontype(T) # singleton types
+            @boundscheck checkbounds(a, i1, tailinds...)
+            return T.instance
+        end
         return reinterpret(T, a.parent[i1, tailinds...])
     else
         @boundscheck checkbounds(a, i1, tailinds...)
         ind_start, sidx = divrem((i1-1)*sizeof(T), sizeof(S))
-        t = Ref{T}()
-        s = Ref{S}()
-        GC.@preserve t s begin
-            tptr = Ptr{UInt8}(unsafe_convert(Ref{T}, t))
-            sptr = Ptr{UInt8}(unsafe_convert(Ref{S}, s))
-            # Optimizations that avoid branches
-            if sizeof(T) % sizeof(S) == 0
-                # T is bigger than S and contains an integer number of them
-                n = sizeof(T) ÷ sizeof(S)
+        # Optimizations that avoid branches
+        if sizeof(T) % sizeof(S) == 0
+            # T is bigger than S and contains an integer number of them
+            n = sizeof(T) ÷ sizeof(S)
+            t = Ref{T}()
+            GC.@preserve t begin
+                sptr = Ptr{S}(unsafe_convert(Ref{T}, t))
                 for i = 1:n
-                    s[] = a.parent[ind_start + i, tailinds...]
-                    _memcpy!(tptr + (i-1)*sizeof(S), sptr, sizeof(S))
+                     s = a.parent[ind_start + i, tailinds...]
+                     unsafe_store!(sptr, s, i)
                 end
-            elseif sizeof(S) % sizeof(T) == 0
-                # S is bigger than T and contains an integer number of them
-                s[] = a.parent[ind_start + 1, tailinds...]
-                _memcpy!(tptr, sptr + sidx, sizeof(T))
-            else
-                i = 1
-                nbytes_copied = 0
-                # This is a bit complicated to deal with partial elements
-                # at both the start and the end. LLVM will fold as appropriate,
-                # once it knows the data layout
+            end
+            return t[]
+        elseif sizeof(S) % sizeof(T) == 0
+            # S is bigger than T and contains an integer number of them
+            s = Ref{S}(a.parent[ind_start + 1, tailinds...])
+            GC.@preserve s begin
+                tptr = Ptr{T}(unsafe_convert(Ref{S}, s))
+                return unsafe_load(tptr + sidx)
+            end
+        else
+            i = 1
+            nbytes_copied = 0
+            # This is a bit complicated to deal with partial elements
+            # at both the start and the end. LLVM will fold as appropriate,
+            # once it knows the data layout
+            s = Ref{S}()
+            t = Ref{T}()
+            GC.@preserve s t begin
+                sptr = Ptr{S}(unsafe_convert(Ref{S}, s))
+                tptr = Ptr{T}(unsafe_convert(Ref{T}, t))
                 while nbytes_copied < sizeof(T)
                     s[] = a.parent[ind_start + i, tailinds...]
                     nb = min(sizeof(S) - sidx, sizeof(T)-nbytes_copied)
@@ -385,61 +440,67 @@ end
                     i += 1
                 end
             end
+            return t[]
         end
-        return t[]
     end
 end
 
 @inline @propagate_inbounds function _getindex_ra(a::ReshapedReinterpretArray{T,N,S}, i1::Int, tailinds::TT) where {T,N,S,TT}
     # Make sure to match the scalar reinterpret if that is applicable
     if sizeof(T) == sizeof(S) && (fieldcount(T) + fieldcount(S)) == 0
+        if issingletontype(T) # singleton types
+            @boundscheck checkbounds(a, i1, tailinds...)
+            return T.instance
+        end
         return reinterpret(T, a.parent[i1, tailinds...])
     end
     @boundscheck checkbounds(a, i1, tailinds...)
     if sizeof(T) >= sizeof(S)
         t = Ref{T}()
-        s = Ref{S}()
-        GC.@preserve t s begin
-            tptr = Ptr{UInt8}(unsafe_convert(Ref{T}, t))
-            sptr = Ptr{UInt8}(unsafe_convert(Ref{S}, s))
+        GC.@preserve t begin
+            sptr = Ptr{S}(unsafe_convert(Ref{T}, t))
             if sizeof(T) > sizeof(S)
                 # Extra dimension in the parent array
                 n = sizeof(T) ÷ sizeof(S)
                 if isempty(tailinds) && IndexStyle(a.parent) === IndexLinear()
                     offset = n * (i1 - firstindex(a))
                     for i = 1:n
-                        s[] = a.parent[i + offset]
-                        _memcpy!(tptr + (i-1)*sizeof(S), sptr, sizeof(S))
+                        s = a.parent[i + offset]
+                        unsafe_store!(sptr, s, i)
                     end
                 else
                     for i = 1:n
-                        s[] = a.parent[i, i1, tailinds...]
-                        _memcpy!(tptr + (i-1)*sizeof(S), sptr, sizeof(S))
+                        s = a.parent[i, i1, tailinds...]
+                        unsafe_store!(sptr, s, i)
                     end
                 end
             else
                 # No extra dimension
-                s[] = a.parent[i1, tailinds...]
-                _memcpy!(tptr, sptr, sizeof(S))
+                s = a.parent[i1, tailinds...]
+                unsafe_store!(sptr, s)
             end
         end
         return t[]
     end
     # S is bigger than T and contains an integer number of them
-    n = sizeof(S) ÷ sizeof(T)
-    t = Ref{NTuple{n,T}}()
+    # n = sizeof(S) ÷ sizeof(T)
     s = Ref{S}()
-    GC.@preserve t s begin
-        tptr = Ptr{UInt8}(unsafe_convert(Ref{T}, t))
-        sptr = Ptr{UInt8}(unsafe_convert(Ref{S}, s))
+    GC.@preserve s begin
+        tptr = Ptr{T}(unsafe_convert(Ref{S}, s))
         s[] = a.parent[tailinds...]
-        _memcpy!(tptr, sptr, sizeof(S))
+        return unsafe_load(tptr, i1)
     end
-    return t[][i1]
 end
 
-@inline @propagate_inbounds setindex!(a::NonReshapedReinterpretArray{T,0,S} where T, v) where {S} = (a.parent[] = reinterpret(S, v))
-@inline @propagate_inbounds setindex!(a::ReinterpretArray, v) = (a[1] = v)
+@inline @propagate_inbounds function setindex!(a::NonReshapedReinterpretArray{T,0,S}, v) where {T,S}
+    if isprimitivetype(S) && isprimitivetype(T)
+        a.parent[] = reinterpret(S, v)
+        return a
+    end
+    setindex!(a, v, firstindex(a))
+end
+
+@inline @propagate_inbounds setindex!(a::ReinterpretArray, v) = setindex!(a, v, firstindex(a))
 
 @inline @propagate_inbounds function setindex!(a::ReinterpretArray{T,N,S}, v, inds::Vararg{Int, N}) where {T,N,S}
     check_writable(a)
@@ -458,12 +519,10 @@ end
 @inline @propagate_inbounds function setindex!(a::ReshapedReinterpretArray{T,N,S}, v, ind::SCartesianIndex2) where {T,N,S}
     check_writable(a)
     v = convert(T, v)::T
-    t = Ref{T}(v)
     s = Ref{S}(a.parent[ind.j])
-    GC.@preserve t s begin
-        tptr = Ptr{UInt8}(unsafe_convert(Ref{T}, t))
-        sptr = Ptr{UInt8}(unsafe_convert(Ref{S}, s))
-        _memcpy!(sptr + (ind.i-1)*sizeof(T), tptr, sizeof(T))
+    GC.@preserve s begin
+        tptr = Ptr{T}(unsafe_convert(Ref{S}, s))
+        unsafe_store!(tptr, v, ind.i)
     end
     a.parent[ind.j] = s[]
     return a
@@ -473,29 +532,41 @@ end
     v = convert(T, v)::T
     # Make sure to match the scalar reinterpret if that is applicable
     if sizeof(T) == sizeof(S) && (fieldcount(T) + fieldcount(S)) == 0
-        return setindex!(a.parent, reinterpret(S, v), i1, tailinds...)
+        if issingletontype(T) # singleton types
+            @boundscheck checkbounds(a, i1, tailinds...)
+            # setindex! is a noop except for the index check
+        else
+            setindex!(a.parent, reinterpret(S, v), i1, tailinds...)
+        end
     else
         @boundscheck checkbounds(a, i1, tailinds...)
         ind_start, sidx = divrem((i1-1)*sizeof(T), sizeof(S))
-        t = Ref{T}(v)
-        s = Ref{S}()
-        GC.@preserve t s begin
-            tptr = Ptr{UInt8}(unsafe_convert(Ref{T}, t))
-            sptr = Ptr{UInt8}(unsafe_convert(Ref{S}, s))
-            # Optimizations that avoid branches
-            if sizeof(T) % sizeof(S) == 0
-                # T is bigger than S and contains an integer number of them
+        # Optimizations that avoid branches
+        if sizeof(T) % sizeof(S) == 0
+            # T is bigger than S and contains an integer number of them
+            t = Ref{T}(v)
+            GC.@preserve t begin
+                sptr = Ptr{S}(unsafe_convert(Ref{T}, t))
                 n = sizeof(T) ÷ sizeof(S)
-                for i = 0:n-1
-                    _memcpy!(sptr, tptr + i*sizeof(S), sizeof(S))
-                    a.parent[ind_start + i + 1, tailinds...] = s[]
+                for i = 1:n
+                    s = unsafe_load(sptr, i)
+                    a.parent[ind_start + i, tailinds...] = s
                 end
-            elseif sizeof(S) % sizeof(T) == 0
-                # S is bigger than T and contains an integer number of them
-                s[] = a.parent[ind_start + 1, tailinds...]
-                _memcpy!(sptr + sidx, tptr, sizeof(T))
+            end
+        elseif sizeof(S) % sizeof(T) == 0
+            # S is bigger than T and contains an integer number of them
+            s = Ref{S}(a.parent[ind_start + 1, tailinds...])
+            GC.@preserve s begin
+                tptr = Ptr{T}(unsafe_convert(Ref{S}, s))
+                unsafe_store!(tptr + sidx, v)
                 a.parent[ind_start + 1, tailinds...] = s[]
-            else
+            end
+        else
+            t = Ref{T}(v)
+            s = Ref{S}()
+            GC.@preserve t s begin
+                tptr = Ptr{UInt8}(unsafe_convert(Ref{T}, t))
+                sptr = Ptr{UInt8}(unsafe_convert(Ref{S}, s))
                 nbytes_copied = 0
                 i = 1
                 # Deal with any partial elements at the start. We'll have to copy in the
@@ -534,39 +605,46 @@ end
     v = convert(T, v)::T
     # Make sure to match the scalar reinterpret if that is applicable
     if sizeof(T) == sizeof(S) && (fieldcount(T) + fieldcount(S)) == 0
-        return setindex!(a.parent, reinterpret(S, v), i1, tailinds...)
+        if issingletontype(T) # singleton types
+            @boundscheck checkbounds(a, i1, tailinds...)
+            # setindex! is a noop except for the index check
+        else
+            setindex!(a.parent, reinterpret(S, v), i1, tailinds...)
+        end
     end
     @boundscheck checkbounds(a, i1, tailinds...)
-    t = Ref{T}(v)
-    s = Ref{S}()
-    GC.@preserve t s begin
-        tptr = Ptr{UInt8}(unsafe_convert(Ref{T}, t))
-        sptr = Ptr{UInt8}(unsafe_convert(Ref{S}, s))
-        if sizeof(T) >= sizeof(S)
+    if sizeof(T) >= sizeof(S)
+        t = Ref{T}(v)
+        GC.@preserve t begin
+            sptr = Ptr{S}(unsafe_convert(Ref{T}, t))
             if sizeof(T) > sizeof(S)
                 # Extra dimension in the parent array
                 n = sizeof(T) ÷ sizeof(S)
                 if isempty(tailinds) && IndexStyle(a.parent) === IndexLinear()
                     offset = n * (i1 - firstindex(a))
                     for i = 1:n
-                        _memcpy!(sptr, tptr + (i-1)*sizeof(S), sizeof(S))
-                        a.parent[i + offset] = s[]
+                        s = unsafe_load(sptr, i)
+                        a.parent[i + offset] = s
                     end
                 else
                     for i = 1:n
-                        _memcpy!(sptr, tptr + (i-1)*sizeof(S), sizeof(S))
-                        a.parent[i, i1, tailinds...] = s[]
+                        s = unsafe_load(sptr, i)
+                        a.parent[i, i1, tailinds...] = s
                     end
                 end
-            else
+            else # sizeof(T) == sizeof(S)
                 # No extra dimension
-                _memcpy!(sptr, tptr, sizeof(S))
-                a.parent[i1, tailinds...] = s[]
+                s = unsafe_load(sptr)
+                a.parent[i1, tailinds...] = s
             end
-        else
-            # S is bigger than T and contains an integer number of them
+        end
+    else
+        # S is bigger than T and contains an integer number of them
+        s = Ref{S}()
+        GC.@preserve s begin
+            tptr = Ptr{T}(unsafe_convert(Ref{S}, s))
             s[] = a.parent[tailinds...]
-            _memcpy!(sptr + (i1-1)*sizeof(T), tptr, sizeof(T))
+            unsafe_store!(tptr, v, i1)
             a.parent[tailinds...] = s[]
         end
     end
@@ -584,7 +662,7 @@ function intersect(p1::Padding, p2::Padding)
     Padding(start, max(0, stop-start))
 end
 
-struct PaddingError
+struct PaddingError <: Exception
     S::Type
     T::Type
 end
@@ -644,25 +722,23 @@ function CyclePadding(T::DataType)
     CyclePadding(pad, as)
 end
 
-using .Iterators: Stateful
-@pure function array_subpadding(S, T)
-    checked_size = 0
+@assume_effects :total function array_subpadding(S, T)
     lcm_size = lcm(sizeof(S), sizeof(T))
-    s, t = Stateful{<:Any, Any}(CyclePadding(S)),
-           Stateful{<:Any, Any}(CyclePadding(T))
+    s, t = CyclePadding(S), CyclePadding(T)
     isempty(t) && return true
     isempty(s) && return false
+    checked_size = 0
+    ps, sstate = iterate(s)   # use of Stateful harms inference and makes this vulnerable to invalidation
+    pad, tstate = iterate(t)
     while checked_size < lcm_size
-        # Take padding in T
-        pad = popfirst!(t)
-        # See if there's corresponding padding in S
         while true
-            ps = peek(s)
+            # See if there's corresponding padding in S
             ps.offset > pad.offset && return false
             intersect(ps, pad) == pad && break
-            popfirst!(s)
+            ps, sstate = iterate(s, sstate)
         end
         checked_size = pad.offset + pad.size
+        pad, tstate = iterate(t, tstate)
     end
     return true
 end
@@ -681,7 +757,7 @@ end
 
 @noinline function mapreduce_impl(f::F, op::OP, A::AbstractArrayOrBroadcasted,
                                   ifirst::SCI, ilast::SCI, blksize::Int) where {F,OP,SCI<:SCartesianIndex2{K}} where K
-    if ifirst.j + blksize > ilast.j
+    if ilast.j - ifirst.j < blksize
         # sequential portion
         @inbounds a1 = A[ifirst]
         @inbounds a2 = A[SCI(2,ifirst.j)]
@@ -700,7 +776,7 @@ end
         return v
     else
         # pairwise portion
-        jmid = (ifirst.j + ilast.j) >> 1
+        jmid = ifirst.j + (ilast.j - ifirst.j) >> 1
         v1 = mapreduce_impl(f, op, A, ifirst, SCI(K,jmid), blksize)
         v2 = mapreduce_impl(f, op, A, SCI(1,jmid+1), ilast, blksize)
         return op(v1, v2)
