@@ -202,12 +202,6 @@ JL_DLLEXPORT void JL_NORETURN jl_eof_error(void)
     jl_throw(jl_new_struct(eof_error));
 }
 
-// get kwsorter field, with appropriate error check and message
-JL_DLLEXPORT jl_value_t *jl_get_keyword_sorter(jl_value_t *f)
-{
-    return jl_get_kwsorter(jl_typeof(f));
-}
-
 JL_DLLEXPORT void jl_typeassert(jl_value_t *x, jl_value_t *t)
 {
     if (!jl_isa(x,t))
@@ -531,7 +525,7 @@ JL_DLLEXPORT jl_value_t *jl_stdout_obj(void) JL_NOTSAFEPOINT
     if (jl_base_module == NULL)
         return NULL;
     jl_binding_t *stdout_obj = jl_get_module_binding(jl_base_module, jl_symbol("stdout"));
-    return stdout_obj ? stdout_obj->value : NULL;
+    return stdout_obj ? jl_atomic_load_relaxed(&stdout_obj->value) : NULL;
 }
 
 JL_DLLEXPORT jl_value_t *jl_stderr_obj(void) JL_NOTSAFEPOINT
@@ -539,7 +533,7 @@ JL_DLLEXPORT jl_value_t *jl_stderr_obj(void) JL_NOTSAFEPOINT
     if (jl_base_module == NULL)
         return NULL;
     jl_binding_t *stderr_obj = jl_get_module_binding(jl_base_module, jl_symbol("stderr"));
-    return stderr_obj ? stderr_obj->value : NULL;
+    return stderr_obj ? jl_atomic_load_relaxed(&stderr_obj->value) : NULL;
 }
 
 // toys for debugging ---------------------------------------------------------
@@ -583,29 +577,29 @@ JL_DLLEXPORT int jl_is_identifier(char *str) JL_NOTSAFEPOINT
     return 1;
 }
 
-static jl_datatype_t *first_arg_datatype(jl_value_t *a JL_PROPAGATES_ROOT, int got_tuple1) JL_NOTSAFEPOINT
+static jl_datatype_t *nth_arg_datatype(jl_value_t *a JL_PROPAGATES_ROOT, int n) JL_NOTSAFEPOINT
 {
     if (jl_is_datatype(a)) {
-        if (got_tuple1)
+        if (n == 0)
             return (jl_datatype_t*)a;
         if (jl_is_tuple_type(a)) {
-            if (jl_nparams(a) < 1)
+            if (jl_nparams(a) < n)
                 return NULL;
-            return first_arg_datatype(jl_tparam0(a), 1);
+            return nth_arg_datatype(jl_tparam(a, n - 1), 0);
         }
         return NULL;
     }
     else if (jl_is_typevar(a)) {
-        return first_arg_datatype(((jl_tvar_t*)a)->ub, got_tuple1);
+        return nth_arg_datatype(((jl_tvar_t*)a)->ub, n);
     }
     else if (jl_is_unionall(a)) {
-        return first_arg_datatype(((jl_unionall_t*)a)->body, got_tuple1);
+        return nth_arg_datatype(((jl_unionall_t*)a)->body, n);
     }
     else if (jl_is_uniontype(a)) {
         jl_uniontype_t *u = (jl_uniontype_t*)a;
-        jl_datatype_t *d1 = first_arg_datatype(u->a, got_tuple1);
+        jl_datatype_t *d1 = nth_arg_datatype(u->a, n);
         if (d1 == NULL) return NULL;
-        jl_datatype_t *d2 = first_arg_datatype(u->b, got_tuple1);
+        jl_datatype_t *d2 = nth_arg_datatype(u->b, n);
         if (d2 == NULL || d1->name != d2->name)
             return NULL;
         return d1;
@@ -614,35 +608,43 @@ static jl_datatype_t *first_arg_datatype(jl_value_t *a JL_PROPAGATES_ROOT, int g
 }
 
 // get DataType of first tuple element (if present), or NULL if cannot be determined
-JL_DLLEXPORT jl_datatype_t *jl_first_argument_datatype(jl_value_t *argtypes JL_PROPAGATES_ROOT) JL_NOTSAFEPOINT
+jl_datatype_t *jl_nth_argument_datatype(jl_value_t *argtypes JL_PROPAGATES_ROOT, int n) JL_NOTSAFEPOINT
 {
-    return first_arg_datatype(argtypes, 0);
+    return nth_arg_datatype(argtypes, n);
 }
 
 // get DataType implied by a single given type, or `nothing`
 JL_DLLEXPORT jl_value_t *jl_argument_datatype(jl_value_t *argt JL_PROPAGATES_ROOT) JL_NOTSAFEPOINT
 {
-    jl_datatype_t *dt = first_arg_datatype(argt, 1);
+    jl_datatype_t *dt = nth_arg_datatype(argt, 0);
     if (dt == NULL)
         return jl_nothing;
     return (jl_value_t*)dt;
 }
 
-static int is_globfunction(jl_value_t *v, jl_datatype_t *dv, jl_sym_t **globname_out)
+static int is_globname_binding(jl_value_t *v, jl_datatype_t *dv) JL_NOTSAFEPOINT
+{
+    jl_sym_t *globname = dv->name->mt != NULL ? dv->name->mt->name : NULL;
+    if (globname && dv->name->module && jl_binding_resolved_p(dv->name->module, globname)) {
+        jl_binding_t *b = jl_get_module_binding(dv->name->module, globname);
+        if (b && b->constp) {
+            jl_value_t *bv = jl_atomic_load_relaxed(&b->value);
+            // The `||` makes this function work for both function instances and function types.
+            if (bv == v || jl_typeof(bv) == v)
+                return 1;
+        }
+    }
+    return 0;
+}
+
+static int is_globfunction(jl_value_t *v, jl_datatype_t *dv, jl_sym_t **globname_out) JL_NOTSAFEPOINT
 {
     jl_sym_t *globname = dv->name->mt != NULL ? dv->name->mt->name : NULL;
     *globname_out = globname;
-    int globfunc = 0;
-    if (globname && !strchr(jl_symbol_name(globname), '#') &&
-        !strchr(jl_symbol_name(globname), '@') && dv->name->module &&
-        jl_binding_resolved_p(dv->name->module, globname)) {
-        jl_binding_t *b = jl_get_module_binding(dv->name->module, globname);
-        // The `||` makes this function work for both function instances and function types.
-        if (b && b->value && (b->value == v || jl_typeof(b->value) == v)) {
-            globfunc = 1;
-        }
+    if (globname && !strchr(jl_symbol_name(globname), '#') && !strchr(jl_symbol_name(globname), '@')) {
+        return 1;
     }
-    return globfunc;
+    return 0;
 }
 
 static size_t jl_static_show_x_sym_escaped(JL_STREAM *out, jl_sym_t *name) JL_NOTSAFEPOINT
@@ -706,6 +708,12 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
         n += jl_static_show_x(out, (jl_value_t*)vt, depth);
         n += jl_printf(out, ">");
     }
+    else if (vt == (jl_datatype_t*)jl_buff_tag) {
+        n += jl_printf(out, "<?#%p::jl_buff_tag marked memory>", (void*)v);
+    }
+    else if (vt == (jl_datatype_t*)(uintptr_t)(0xbabababababababaull & ~15)) {
+        n += jl_printf(out, "<?#%p::baaaaaad>", (void*)v);
+    }
     // These need to be special cased because they
     // exist only by pointer identity in early startup
     else if (v == (jl_value_t*)jl_simplevector_type) {
@@ -773,10 +781,41 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
         // `Base.Set{Int}`, and function types are printed as e.g. `typeof(Main.f)`
         jl_datatype_t *dv = (jl_datatype_t*)v;
         jl_sym_t *globname;
-        int globfunc = is_globfunction(v, dv, &globname);
+        int globfunc = is_globname_binding(v, dv) && is_globfunction(v, dv, &globname);
         jl_sym_t *sym = globfunc ? globname : dv->name->name;
         char *sn = jl_symbol_name(sym);
         size_t quote = 0;
+        if (dv->name == jl_tuple_typename) {
+            if (dv == jl_tuple_type)
+                return jl_printf(out, "Tuple");
+            int taillen = 1, tlen = jl_nparams(dv), i;
+            for (i = tlen-2; i >= 0; i--) {
+                if (jl_tparam(dv, i) == jl_tparam(dv, tlen-1))
+                    taillen++;
+                else
+                    break;
+            }
+            if (taillen == tlen && taillen > 3) {
+                n += jl_printf(out, "NTuple{%d, ", tlen);
+                n += jl_static_show_x(out, jl_tparam0(dv), depth);
+                n += jl_printf(out, "}");
+            }
+            else {
+                n += jl_printf(out, "Tuple{");
+                for (i = 0; i < (taillen > 3 ? tlen-taillen : tlen); i++) {
+                    if (i > 0)
+                        n += jl_printf(out, ", ");
+                    n += jl_static_show_x(out, jl_tparam(dv, i), depth);
+                }
+                if (taillen > 3) {
+                    n += jl_printf(out, ", Vararg{");
+                    n += jl_static_show_x(out, jl_tparam(dv, tlen-1), depth);
+                    n += jl_printf(out, ", %d}", taillen);
+                }
+                n += jl_printf(out, "}");
+            }
+            return n;
+        }
         if (globfunc) {
             n += jl_printf(out, "typeof(");
         }
@@ -796,9 +835,7 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
                 n += jl_printf(out, ")");
             }
         }
-        if (dv->parameters && (jl_value_t*)dv != dv->name->wrapper &&
-            (jl_has_free_typevars(v) ||
-             (jl_value_t*)dv != (jl_value_t*)jl_tuple_type)) {
+        if (dv->parameters && (jl_value_t*)dv != dv->name->wrapper) {
             size_t j, tlen = jl_nparams(dv);
             if (tlen > 0) {
                 n += jl_printf(out, "{");
@@ -810,9 +847,6 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
                 }
                 n += jl_printf(out, "}");
             }
-            else if (dv->name == jl_tuple_typename) {
-                n += jl_printf(out, "{}");
-            }
         }
     }
     else if (vt == jl_intrinsic_type) {
@@ -1065,20 +1099,18 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
         n += jl_static_show_x(out, *(jl_value_t**)v, depth);
         n += jl_printf(out, ")");
     }
-    else if (jl_static_is_function_(vt)) {
+    else if (jl_static_is_function_(vt) && is_globname_binding(v, (jl_datatype_t*)vt)) {
         // v is function instance (an instance of a Function type).
         jl_datatype_t *dv = (jl_datatype_t*)vt;
-        jl_sym_t *sym = dv->name->mt->name;
-        char *sn = jl_symbol_name(sym);
-
-        jl_sym_t *globname;
-        int globfunc = is_globfunction(v, dv, &globname);
+        jl_sym_t *sym;
+        int globfunc = is_globfunction(v, dv, &sym);
         int quote = 0;
         if (jl_core_module && (dv->name->module != jl_core_module || !jl_module_exports_p(jl_core_module, sym))) {
             n += jl_static_show_x(out, (jl_value_t*)dv->name->module, depth);
             n += jl_printf(out, ".");
 
             size_t i = 0;
+            char *sn = jl_symbol_name(sym);
             if (globfunc && !jl_id_start_char(u8_nextchar(sn, &i))) {
                 n += jl_printf(out, ":(");
                 quote = 1;
@@ -1144,7 +1176,8 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
             }
             if (vt == jl_typemap_entry_type) {
                 n += jl_printf(out, ", next=↩︎\n  ");
-                n += jl_static_show_next_(out, (jl_value_t*)((jl_typemap_entry_t*)v)->next, v, depth);
+                jl_value_t *next = (jl_value_t*)jl_atomic_load_relaxed(&((jl_typemap_entry_t*)v)->next);
+                n += jl_static_show_next_(out, next, v, depth);
             }
         }
         n += jl_printf(out, ")");
@@ -1193,12 +1226,12 @@ static size_t jl_static_show_next_(JL_STREAM *out, jl_value_t *v, jl_value_t *pr
                 }
                 // verify that we aren't trying to follow a circular list
                 // by following the list again, and ensuring this is the only link to next
-                jl_value_t *mnext = (jl_value_t*)((jl_typemap_entry_t*)m)->next;
+                jl_value_t *mnext = (jl_value_t*)jl_atomic_load_relaxed(&((jl_typemap_entry_t*)m)->next);
                 jl_value_t *m2 = p->v;
                 if (m2 == mnext)
                     break;
                 while (m2 && jl_typeis(m2, jl_typemap_entry_type)) {
-                    jl_value_t *mnext2 = (jl_value_t*)((jl_typemap_entry_t*)m2)->next;
+                    jl_value_t *mnext2 = (jl_value_t*)jl_atomic_load_relaxed(&((jl_typemap_entry_t*)m2)->next);
                     if (mnext2 == mnext) {
                         if (m2 != m)
                             mnext = NULL;
@@ -1227,7 +1260,7 @@ JL_DLLEXPORT size_t jl_static_show_func_sig(JL_STREAM *s, jl_value_t *type) JL_N
 {
     size_t n = 0;
     size_t i;
-    jl_value_t *ftype = (jl_value_t*)jl_first_argument_datatype(type);
+    jl_value_t *ftype = (jl_value_t*)jl_nth_argument_datatype(type, 1);
     if (ftype == NULL)
         return jl_static_show(s, type);
     jl_unionall_t *tvars = (jl_unionall_t*)type;
@@ -1246,7 +1279,9 @@ JL_DLLEXPORT size_t jl_static_show_func_sig(JL_STREAM *s, jl_value_t *type) JL_N
         n += jl_static_show(s, type);
         return n;
     }
-    if (jl_nparams(ftype) == 0 || ftype == ((jl_datatype_t*)ftype)->name->wrapper) {
+    if ((jl_nparams(ftype) == 0 || ftype == ((jl_datatype_t*)ftype)->name->wrapper) &&
+            ((jl_datatype_t*)ftype)->name->mt != jl_type_type_mt &&
+            ((jl_datatype_t*)ftype)->name->mt != jl_nonfunction_mt) {
         n += jl_printf(s, "%s", jl_symbol_name(((jl_datatype_t*)ftype)->name->mt->name));
     }
     else {
@@ -1316,6 +1351,11 @@ JL_DLLEXPORT void jl_breakpoint(jl_value_t *v)
     // put a breakpoint in your debugger here
 }
 
+JL_DLLEXPORT void jl_test_failure_breakpoint(jl_value_t *v)
+{
+    // put a breakpoint in your debugger here
+}
+
 // logging tools --------------------------------------------------------------
 
 void jl_log(int level, jl_value_t *module, jl_value_t *group, jl_value_t *id,
