@@ -9,6 +9,7 @@
 #include <mach/task.h>
 #include <mach/mig_errors.h>
 #include <AvailabilityMacros.h>
+#include "mach_excServer.c"
 
 #ifdef MAC_OS_X_VERSION_10_9
 #include <sys/_types/_ucontext64.h>
@@ -40,6 +41,8 @@ static void attach_exception_port(thread_port_t thread, int segv_only);
 
 // low 16 bits are the thread id, the next 8 bits are the original gc_state
 static arraylist_t suspended_threads;
+extern uv_mutex_t safepoint_lock;
+extern uv_cond_t safepoint_cond;
 void jl_mach_gc_end(void)
 {
     // Requires the safepoint lock to be held
@@ -47,7 +50,7 @@ void jl_mach_gc_end(void)
         uintptr_t item = (uintptr_t)suspended_threads.items[i];
         int16_t tid = (int16_t)item;
         int8_t gc_state = (int8_t)(item >> 8);
-        jl_ptls_t ptls2 = jl_all_tls_states[tid];
+        jl_ptls_t ptls2 = jl_atomic_load_relaxed(&jl_all_tls_states)[tid];
         jl_atomic_store_release(&ptls2->gc_state, gc_state);
         thread_resume(pthread_mach_thread_np(ptls2->system_id));
     }
@@ -68,6 +71,14 @@ static int jl_mach_gc_wait(jl_ptls_t ptls2,
         return 0;
     }
     // Otherwise, set the gc state of the thread, suspend and record it
+    // TODO: TSAN will complain that it never saw the faulting task do an
+    // atomic release (it was in the kernel). And our attempt here does
+    // nothing, since we are a different thread, and it is not transitive).
+    //
+    // This also means we are not making this thread available for GC work.
+    // Eventually, we should probably release this signal to the original
+    // thread, (return KERN_FAILURE instead of KERN_SUCCESS) so that it
+    // triggers a SIGSEGV and gets handled by the usual codepath for unix.
     int8_t gc_state = ptls2->gc_state;
     jl_atomic_store_release(&ptls2->gc_state, JL_GC_STATE_WAITING);
     uintptr_t item = tid | (((uintptr_t)gc_state) << 16);
@@ -79,21 +90,17 @@ static int jl_mach_gc_wait(jl_ptls_t ptls2,
 
 static mach_port_t segv_port = 0;
 
-extern boolean_t exc_server(mach_msg_header_t *, mach_msg_header_t *);
-
 #define STR(x) #x
 #define XSTR(x) STR(x)
 #define HANDLE_MACH_ERROR(msg, retval) \
-    if (retval != KERN_SUCCESS) { mach_error(msg XSTR(: __FILE__:__LINE__:), (retval)); jl_exit(1); }
+    if (retval != KERN_SUCCESS) { mach_error(msg XSTR(: __FILE__:__LINE__:), (retval)); abort(); }
 
 void *mach_segv_listener(void *arg)
 {
     (void)arg;
-    while (1) {
-        int ret = mach_msg_server(exc_server, 2048, segv_port, MACH_MSG_TIMEOUT_NONE);
-        jl_safe_printf("mach_msg_server: %s\n", mach_error_string(ret));
-        jl_exit(128 + SIGSEGV);
-    }
+    int ret = mach_msg_server(mach_exc_server, 2048, segv_port, MACH_MSG_TIMEOUT_NONE);
+    mach_error("mach_msg_server" XSTR(: __FILE__:__LINE__:), ret);
+    abort();
 }
 
 
@@ -110,7 +117,8 @@ static void allocate_mach_handler()
     if (_keymgr_set_lockmode_processwide_ptr(KEYMGR_GCC3_DW2_OBJ_LIST, NM_ALLOW_RECURSION))
         jl_error("_keymgr_set_lockmode_processwide_ptr failed");
 
-    arraylist_new(&suspended_threads, jl_n_threads);
+    int16_t nthreads = jl_atomic_load_acquire(&jl_n_threads);
+    arraylist_new(&suspended_threads, nthreads); // we will resize later (inside safepoint_lock), if needed
     pthread_t thread;
     pthread_attr_t attr;
     kern_return_t ret;
@@ -128,35 +136,32 @@ static void allocate_mach_handler()
         jl_error("pthread_create failed");
     }
     pthread_attr_destroy(&attr);
-    for (int16_t tid = 0; tid < jl_n_threads; tid++) {
-        attach_exception_port(pthread_mach_thread_np(jl_all_tls_states[tid]->system_id), 0);
-    }
 }
 
 #ifdef LLVMLIBUNWIND
 volatile mach_port_t mach_profiler_thread = 0;
-static kern_return_t profiler_segv_handler
-                (mach_port_t                          exception_port,
-                 mach_port_t                                  thread,
-                 mach_port_t                                    task,
-                 exception_type_t                          exception,
-                 exception_data_t                               code,
-                 mach_msg_type_number_t                   code_count);
+static kern_return_t profiler_segv_handler(
+    mach_port_t exception_port,
+    mach_port_t thread,
+    mach_port_t task,
+    exception_type_t exception,
+    mach_exception_data_t code,
+    mach_msg_type_number_t codeCnt);
 #endif
 
 #if defined(_CPU_X86_64_)
 typedef x86_thread_state64_t host_thread_state_t;
 typedef x86_exception_state64_t host_exception_state_t;
-#define THREAD_STATE x86_THREAD_STATE64
-#define THREAD_STATE_COUNT x86_THREAD_STATE64_COUNT
+#define MACH_THREAD_STATE x86_THREAD_STATE64
+#define MACH_THREAD_STATE_COUNT x86_THREAD_STATE64_COUNT
 #define HOST_EXCEPTION_STATE x86_EXCEPTION_STATE64
 #define HOST_EXCEPTION_STATE_COUNT x86_EXCEPTION_STATE64_COUNT
 
 #elif defined(_CPU_AARCH64_)
 typedef arm_thread_state64_t host_thread_state_t;
 typedef arm_exception_state64_t host_exception_state_t;
-#define THREAD_STATE ARM_THREAD_STATE64
-#define THREAD_STATE_COUNT ARM_THREAD_STATE64_COUNT
+#define MACH_THREAD_STATE ARM_THREAD_STATE64
+#define MACH_THREAD_STATE_COUNT ARM_THREAD_STATE64_COUNT
 #define HOST_EXCEPTION_STATE ARM_EXCEPTION_STATE64
 #define HOST_EXCEPTION_STATE_COUNT ARM_EXCEPTION_STATE64_COUNT
 #endif
@@ -207,14 +212,13 @@ int is_write_fault(host_exception_state_t exc_state) {
 }
 #endif
 
-static void jl_throw_in_thread(int tid, mach_port_t thread, jl_value_t *exception)
+static void jl_throw_in_thread(jl_ptls_t ptls2, mach_port_t thread, jl_value_t *exception)
 {
-    unsigned int count = THREAD_STATE_COUNT;
+    unsigned int count = MACH_THREAD_STATE_COUNT;
     host_thread_state_t state;
-    kern_return_t ret = thread_get_state(thread, THREAD_STATE, (thread_state_t)&state, &count);
+    kern_return_t ret = thread_get_state(thread, MACH_THREAD_STATE, (thread_state_t)&state, &count);
     HANDLE_MACH_ERROR("thread_get_state", ret);
-    jl_ptls_t ptls2 = jl_all_tls_states[tid];
-    if (!jl_get_safe_restore()) {
+    if (1) { // XXX: !jl_has_safe_restore(ptls2)
         assert(exception);
         ptls2->bt_size =
             rec_backtrace_ctx(ptls2->bt_data, JL_MAX_BT_SIZE, (bt_context_t *)&state,
@@ -222,7 +226,7 @@ static void jl_throw_in_thread(int tid, mach_port_t thread, jl_value_t *exceptio
         ptls2->sig_exception = exception;
     }
     jl_call_in_state(ptls2, &state, &jl_sig_throw);
-    ret = thread_set_state(thread, THREAD_STATE, (thread_state_t)&state, count);
+    ret = thread_set_state(thread, MACH_THREAD_STATE, (thread_state_t)&state, count);
     HANDLE_MACH_ERROR("thread_set_state", ret);
 }
 
@@ -239,51 +243,70 @@ static void segv_handler(int sig, siginfo_t *info, void *context)
     }
 }
 
-//exc_server uses dlsym to find symbol
-JL_DLLEXPORT
-kern_return_t catch_exception_raise(mach_port_t            exception_port,
-                                    mach_port_t            thread,
-                                    mach_port_t            task,
-                                    exception_type_t       exception,
-                                    exception_data_t       code,
-                                    mach_msg_type_number_t code_count)
+// n.b. mach_exc_server expects us to define this symbol locally
+/* The documentation for catch_exception_raise says: A return value of
+ * KERN_SUCCESS indicates that the thread is to continue from the point of
+ * exception. A return value of MIG_NO_REPLY indicates that the exception was
+ * handled directly and the thread was restarted or terminated by the exception
+ * handler. A return value of MIG_DESTROY_REQUEST causes the kernel to try
+ * another exception handler (or terminate the thread). Any other value will
+ * cause mach_msg_server to remove the task and thread port references.
+ *
+ * However MIG_DESTROY_REQUEST does not exist, not does it appear the source
+ * code for mach_msg_server ever destroy those references (only the message
+ * itself).
+ */
+kern_return_t catch_mach_exception_raise(
+    mach_port_t exception_port,
+    mach_port_t thread,
+    mach_port_t task,
+    exception_type_t exception,
+    mach_exception_data_t code,
+    mach_msg_type_number_t codeCnt)
 {
     unsigned int exc_count = HOST_EXCEPTION_STATE_COUNT;
     host_exception_state_t exc_state;
 #ifdef LLVMLIBUNWIND
     if (thread == mach_profiler_thread) {
-        return profiler_segv_handler(exception_port, thread, task, exception, code, code_count);
+        return profiler_segv_handler(exception_port, thread, task, exception, code, codeCnt);
     }
 #endif
     int16_t tid;
     jl_ptls_t ptls2 = NULL;
-    for (tid = 0; tid < jl_n_threads; tid++) {
-        jl_ptls_t _ptls2 = jl_all_tls_states[tid];
+    int nthreads = jl_atomic_load_acquire(&jl_n_threads);
+    for (tid = 0; tid < nthreads; tid++) {
+        jl_ptls_t _ptls2 = jl_atomic_load_relaxed(&jl_all_tls_states)[tid];
         if (pthread_mach_thread_np(_ptls2->system_id) == thread) {
             ptls2 = _ptls2;
             break;
         }
     }
-    if (!ptls2) {
+    if (!ptls2 || ptls2->current_task == NULL) {
         // We don't know about this thread, let the kernel try another handler
         // instead. This shouldn't actually happen since we only register the
         // handler for the threads we know about.
         jl_safe_printf("ERROR: Exception handler triggered on unmanaged thread.\n");
         return KERN_INVALID_ARGUMENT;
     }
+    // XXX: jl_throw_in_thread or segv_handler will eventually check this, but
+    //      we would like to avoid some of this work if we could detect this earlier
+    // if (jl_has_safe_restore(ptls2)) {
+    //     jl_throw_in_thread(ptls2, thread, jl_stackovf_exception);
+    //     return KERN_SUCCESS;
+    // }
+    if (ptls2->gc_state == JL_GC_STATE_WAITING)
+        return KERN_FAILURE;
     if (exception == EXC_ARITHMETIC) {
-        jl_throw_in_thread(tid, thread, jl_diverror_exception);
+        jl_throw_in_thread(ptls2, thread, jl_diverror_exception);
         return KERN_SUCCESS;
     }
-    assert(exception == EXC_BAD_ACCESS);
+    assert(exception == EXC_BAD_ACCESS); // SIGSEGV or SIGBUS
+    if (codeCnt < 2 || code[0] != KERN_PROTECTION_FAILURE) // SEGV_ACCERR or BUS_ADRERR or BUS_ADRALN
+        return KERN_FAILURE;
+    uint64_t fault_addr = code[1];
     kern_return_t ret = thread_get_state(thread, HOST_EXCEPTION_STATE, (thread_state_t)&exc_state, &exc_count);
     HANDLE_MACH_ERROR("thread_get_state", ret);
-#ifdef _CPU_X86_64_
-    uint64_t fault_addr = exc_state.__faultvaddr;
-#else
-    uint64_t fault_addr = exc_state.__far;
-#endif
-    if (jl_addr_is_safepoint(fault_addr)) {
+    if (jl_addr_is_safepoint(fault_addr) && !is_write_fault(exc_state)) {
         if (jl_mach_gc_wait(ptls2, thread, tid))
             return KERN_SUCCESS;
         if (ptls2->tid != 0)
@@ -293,42 +316,54 @@ kern_return_t catch_exception_raise(mach_port_t            exception_port,
         }
         else if (jl_safepoint_consume_sigint()) {
             jl_clear_force_sigint();
-            jl_throw_in_thread(tid, thread, jl_interrupt_exception);
+            jl_throw_in_thread(ptls2, thread, jl_interrupt_exception);
         }
         return KERN_SUCCESS;
     }
-    if (jl_get_safe_restore()) {
-        jl_throw_in_thread(tid, thread, jl_stackovf_exception);
-        return KERN_SUCCESS;
+    if (ptls2->current_task->eh == NULL)
+        return KERN_FAILURE;
+    jl_value_t *excpt;
+    if (is_addr_on_stack(jl_atomic_load_relaxed(&ptls2->current_task), (void*)fault_addr)) {
+        excpt = jl_stackovf_exception;
     }
-#ifdef SEGV_EXCEPTION
-    if (1) {
-#else
-    if (msync((void*)(fault_addr & ~(jl_page_size - 1)), 1, MS_ASYNC) == 0) { // check if this was a valid address
-#endif
-        jl_value_t *excpt;
-        if (is_addr_on_stack(jl_atomic_load_relaxed(&ptls2->current_task), (void*)fault_addr)) {
-            excpt = jl_stackovf_exception;
-        }
-#ifdef SEGV_EXCEPTION
-        else if (msync((void*)(fault_addr & ~(jl_page_size - 1)), 1, MS_ASYNC) != 0) {
-            // no page mapped at this address
-            excpt = jl_segv_exception;
-        }
-#endif
-        else {
-            if (!is_write_fault(exc_state))
-                return KERN_INVALID_ARGUMENT;
-            excpt = jl_readonlymemory_exception;
-        }
-        jl_throw_in_thread(tid, thread, excpt);
+    else if (is_write_fault(exc_state)) // false for alignment errors
+        excpt = jl_readonlymemory_exception;
+    else
+        return KERN_FAILURE;
+    jl_throw_in_thread(ptls2, thread, excpt);
+    return KERN_SUCCESS;
+}
 
-        return KERN_SUCCESS;
-    }
-    else {
-        jl_exit_thread0(128 + SIGSEGV, NULL, 0);
-        return KERN_SUCCESS;
-    }
+//mach_exc_server expects us to define this symbol locally
+kern_return_t catch_mach_exception_raise_state(
+    mach_port_t exception_port,
+    exception_type_t exception,
+    const mach_exception_data_t code,
+    mach_msg_type_number_t codeCnt,
+    int *flavor,
+    const thread_state_t old_state,
+    mach_msg_type_number_t old_stateCnt,
+    thread_state_t new_state,
+    mach_msg_type_number_t *new_stateCnt)
+{
+    return KERN_INVALID_ARGUMENT; // we only use EXCEPTION_DEFAULT
+}
+
+//mach_exc_server expects us to define this symbol locally
+kern_return_t catch_mach_exception_raise_state_identity(
+    mach_port_t exception_port,
+    mach_port_t thread,
+    mach_port_t task,
+    exception_type_t exception,
+    mach_exception_data_t code,
+    mach_msg_type_number_t codeCnt,
+    int *flavor,
+    thread_state_t old_state,
+    mach_msg_type_number_t old_stateCnt,
+    thread_state_t new_state,
+    mach_msg_type_number_t *new_stateCnt)
+{
+    return KERN_INVALID_ARGUMENT; // we only use EXCEPTION_DEFAULT
 }
 
 static void attach_exception_port(thread_port_t thread, int segv_only)
@@ -338,36 +373,47 @@ static void attach_exception_port(thread_port_t thread, int segv_only)
     exception_mask_t mask = EXC_MASK_BAD_ACCESS;
     if (!segv_only)
         mask |= EXC_MASK_ARITHMETIC;
-    ret = thread_set_exception_ports(thread, mask, segv_port, EXCEPTION_DEFAULT, MACHINE_THREAD_STATE);
+    ret = thread_set_exception_ports(thread, mask, segv_port, EXCEPTION_DEFAULT | MACH_EXCEPTION_CODES, MACH_THREAD_STATE);
     HANDLE_MACH_ERROR("thread_set_exception_ports", ret);
 }
 
-static void jl_thread_suspend_and_get_state2(int tid, host_thread_state_t *ctx)
+static int jl_thread_suspend_and_get_state2(int tid, host_thread_state_t *ctx)
 {
-    jl_ptls_t ptls2 = jl_all_tls_states[tid];
+    jl_ptls_t ptls2 = jl_atomic_load_relaxed(&jl_all_tls_states)[tid];
+    if (ptls2 == NULL) // this thread is not alive
+        return 0;
+    jl_task_t *ct2 = ptls2 ? jl_atomic_load_relaxed(&ptls2->current_task) : NULL;
+    if (ct2 == NULL) // this thread is already dead
+        return 0;
+
     mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
 
     kern_return_t ret = thread_suspend(thread);
     HANDLE_MACH_ERROR("thread_suspend", ret);
 
     // Do the actual sampling
-    unsigned int count = THREAD_STATE_COUNT;
+    unsigned int count = MACH_THREAD_STATE_COUNT;
     memset(ctx, 0, sizeof(*ctx));
 
     // Get the state of the suspended thread
-    ret = thread_get_state(thread, THREAD_STATE, (thread_state_t)ctx, &count);
+    ret = thread_get_state(thread, MACH_THREAD_STATE, (thread_state_t)ctx, &count);
+    return 1;
 }
 
-static void jl_thread_suspend_and_get_state(int tid, unw_context_t **ctx)
+static void jl_thread_suspend_and_get_state(int tid, int timeout, unw_context_t **ctx)
 {
+    (void)timeout;
     static host_thread_state_t state;
-    jl_thread_suspend_and_get_state2(tid, &state);
+    if (!jl_thread_suspend_and_get_state2(tid, &state)) {
+        *ctx = NULL;
+        return;
+    }
     *ctx = (unw_context_t*)&state;
 }
 
 static void jl_thread_resume(int tid, int sig)
 {
-    jl_ptls_t ptls2 = jl_all_tls_states[tid];
+    jl_ptls_t ptls2 = jl_atomic_load_relaxed(&jl_all_tls_states)[tid];
     mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
     kern_return_t ret = thread_resume(thread);
     HANDLE_MACH_ERROR("thread_resume", ret);
@@ -377,7 +423,7 @@ static void jl_thread_resume(int tid, int sig)
 // or if SIGINT happens too often.
 static void jl_try_deliver_sigint(void)
 {
-    jl_ptls_t ptls2 = jl_all_tls_states[0];
+    jl_ptls_t ptls2 = jl_atomic_load_relaxed(&jl_all_tls_states)[0];
     mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
 
     kern_return_t ret = thread_suspend(thread);
@@ -394,7 +440,7 @@ static void jl_try_deliver_sigint(void)
         if (force)
             jl_safe_printf("WARNING: Force throwing a SIGINT\n");
         jl_clear_force_sigint();
-        jl_throw_in_thread(0, thread, jl_interrupt_exception);
+        jl_throw_in_thread(ptls2, thread, jl_interrupt_exception);
     }
     else {
         jl_wake_libuv();
@@ -404,58 +450,43 @@ static void jl_try_deliver_sigint(void)
     HANDLE_MACH_ERROR("thread_resume", ret);
 }
 
-static void JL_NORETURN jl_exit_thread0_cb(int exitstate)
+static void JL_NORETURN jl_exit_thread0_cb(int signo)
 {
 CFI_NORETURN
-    jl_critical_error(exitstate - 128, NULL, jl_current_task);
-    jl_exit(exitstate);
+    jl_critical_error(signo, 0, NULL, jl_current_task);
+    jl_atexit_hook(128);
+    jl_raise(signo);
 }
 
-static void jl_exit_thread0(int exitstate, jl_bt_element_t *bt_data, size_t bt_size)
+static void jl_exit_thread0(int signo, jl_bt_element_t *bt_data, size_t bt_size)
 {
-    jl_ptls_t ptls2 = jl_all_tls_states[0];
+    jl_ptls_t ptls2 = jl_atomic_load_relaxed(&jl_all_tls_states)[0];
     mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
 
     host_thread_state_t state;
-    jl_thread_suspend_and_get_state2(0, &state);
-    unw_context_t *uc = (unw_context_t*)&state;
+    if (!jl_thread_suspend_and_get_state2(0, &state)) {
+        // thread 0 is gone? just do the signal ourself
+        jl_raise(signo);
+    }
 
     // This aborts `sleep` and other syscalls.
     kern_return_t ret = thread_abort(thread);
     HANDLE_MACH_ERROR("thread_abort", ret);
 
-    if (bt_data == NULL) {
-        // Must avoid extended backtrace frames here unless we're sure bt_data
-        // is properly rooted.
-        ptls2->bt_size = rec_backtrace_ctx(ptls2->bt_data, JL_MAX_BT_SIZE, uc, NULL);
-    }
-    else {
-        ptls2->bt_size = bt_size; // <= JL_MAX_BT_SIZE
-        memcpy(ptls2->bt_data, bt_data, ptls2->bt_size * sizeof(bt_data[0]));
-    }
-
-    void (*exit_func)(int) = &_exit;
-    if (thread0_exit_count <= 1) {
-        exit_func = &jl_exit_thread0_cb;
-    }
-    else if (thread0_exit_count == 2) {
-        exit_func = &exit;
-    }
-    else {
-        exit_func = &_exit;
-    }
+    ptls2->bt_size = bt_size; // <= JL_MAX_BT_SIZE
+    memcpy(ptls2->bt_data, bt_data, ptls2->bt_size * sizeof(bt_data[0]));
 
 #ifdef _CPU_X86_64_
     // First integer argument. Not portable but good enough =)
-    state.__rdi = exitstate;
+    state.__rdi = signo;
 #elif defined(_CPU_AARCH64_)
-    state.__x[0] = exitstate;
+    state.__x[0] = signo;
 #else
 #error Fill in first integer argument here
 #endif
-    jl_call_in_state(ptls2, &state, (void (*)(void))exit_func);
-    unsigned int count = THREAD_STATE_COUNT;
-    ret = thread_set_state(thread, THREAD_STATE, (thread_state_t)&state, count);
+    jl_call_in_state(ptls2, &state, (void (*)(void))&jl_exit_thread0_cb);
+    unsigned int count = MACH_THREAD_STATE_COUNT;
+    ret = thread_set_state(thread, MACH_THREAD_STATE, (thread_state_t)&state, count);
     HANDLE_MACH_ERROR("thread_set_state", ret);
 
     ret = thread_resume(thread);
@@ -472,29 +503,29 @@ static mach_port_t profile_port = 0;
 volatile static int forceDwarf = -2;
 static unw_context_t profiler_uc;
 
-static kern_return_t profiler_segv_handler
-                (mach_port_t                          exception_port,
-                 mach_port_t                                  thread,
-                 mach_port_t                                    task,
-                 exception_type_t                          exception,
-                 exception_data_t                               code,
-                 mach_msg_type_number_t                   code_count)
+static kern_return_t profiler_segv_handler(
+    mach_port_t exception_port,
+    mach_port_t thread,
+    mach_port_t task,
+    exception_type_t exception,
+    mach_exception_data_t code,
+    mach_msg_type_number_t codeCnt)
 {
     assert(thread == mach_profiler_thread);
     host_thread_state_t state;
 
     // Not currently unwinding. Raise regular segfault
     if (forceDwarf == -2)
-        return KERN_INVALID_ARGUMENT;
+        return KERN_FAILURE;
 
     if (forceDwarf == 0)
         forceDwarf = 1;
     else
         forceDwarf = -1;
 
-    unsigned int count = THREAD_STATE_COUNT;
+    unsigned int count = MACH_THREAD_STATE_COUNT;
 
-    thread_get_state(thread, THREAD_STATE, (thread_state_t)&state, &count);
+    thread_get_state(thread, MACH_THREAD_STATE, (thread_state_t)&state, &count);
 
 #ifdef _CPU_X86_64_
     // don't change cs fs gs rflags
@@ -519,7 +550,7 @@ static kern_return_t profiler_segv_handler
     state.__cpsr = cpsr;
 #endif
 
-    kern_return_t ret = thread_set_state(thread, THREAD_STATE, (thread_state_t)&state, count);
+    kern_return_t ret = thread_set_state(thread, MACH_THREAD_STATE, (thread_state_t)&state, count);
     HANDLE_MACH_ERROR("thread_set_state", ret);
 
     return KERN_SUCCESS;
@@ -568,10 +599,12 @@ void *mach_profile_listener(void *arg)
         // sample each thread, round-robin style in reverse order
         // (so that thread zero gets notified last)
         int keymgr_locked = jl_lock_profile_mach(0);
-        jl_shuffle_int_array_inplace(profile_round_robin_thread_order, jl_n_threads, &profile_cong_rng_seed);
-        for (int idx = jl_n_threads; idx-- > 0; ) {
-            // Stop the threads in the random round-robin order.
-            int i = profile_round_robin_thread_order[idx];
+
+        int nthreads = jl_atomic_load_acquire(&jl_n_threads);
+        int *randperm = profile_get_randperm(nthreads);
+        for (int idx = nthreads; idx-- > 0; ) {
+            // Stop the threads in the random or reverse round-robin order.
+            int i = randperm[idx];
             // if there is no space left, break early
             if (jl_profile_is_buffer_full()) {
                 jl_profile_stop_timer();
@@ -581,7 +614,8 @@ void *mach_profile_listener(void *arg)
             if (_dyld_atfork_prepare != NULL && _dyld_atfork_parent != NULL)
                 _dyld_atfork_prepare(); // briefly acquire the dlsym lock
             host_thread_state_t state;
-            jl_thread_suspend_and_get_state2(i, &state);
+            if (!jl_thread_suspend_and_get_state2(i, &state))
+                continue;
             unw_context_t *uc = (unw_context_t*)&state;
             if (_dyld_atfork_prepare != NULL && _dyld_atfork_parent != NULL)
                 _dyld_atfork_parent(); // quickly release the dlsym lock
@@ -620,12 +654,12 @@ void *mach_profile_listener(void *arg)
 #else
                 bt_size_cur += rec_backtrace_ctx((jl_bt_element_t*)bt_data_prof + bt_size_cur, bt_size_max - bt_size_cur - 1, uc, NULL);
 #endif
-                jl_ptls_t ptls = jl_all_tls_states[i];
+                jl_ptls_t ptls = jl_atomic_load_relaxed(&jl_all_tls_states)[i];
 
                 // store threadid but add 1 as 0 is preserved to indicate end of block
                 bt_data_prof[bt_size_cur++].uintptr = ptls->tid + 1;
 
-                // store task id
+                // store task id (never null)
                 bt_data_prof[bt_size_cur++].jlvalue = (jl_value_t*)jl_atomic_load_relaxed(&ptls->current_task);
 
                 // store cpu cycle clock
