@@ -36,18 +36,26 @@ extern "C" {
 
 #include "threading.h"
 
+JL_DLLEXPORT _Atomic(uint8_t) jl_measure_compile_time_enabled = 0;
+JL_DLLEXPORT _Atomic(uint64_t) jl_cumulative_compile_time = 0;
+JL_DLLEXPORT _Atomic(uint64_t) jl_cumulative_recompile_time = 0;
+
 JL_DLLEXPORT void *jl_get_ptls_states(void)
 {
     // mostly deprecated: use current_task instead
     return jl_current_task->ptls;
 }
 
+static void jl_delete_thread(void*);
+
 #if !defined(_OS_WINDOWS_)
+static pthread_key_t jl_task_exit_key;
 static pthread_key_t jl_safe_restore_key;
 
 __attribute__((constructor)) void _jl_init_safe_restore(void)
 {
     pthread_key_create(&jl_safe_restore_key, NULL);
+    pthread_key_create(&jl_task_exit_key, jl_delete_thread);
 }
 
 JL_DLLEXPORT jl_jmp_buf *jl_get_safe_restore(void)
@@ -120,21 +128,26 @@ static DWORD jl_safe_restore_key;
 BOOLEAN WINAPI DllMain(IN HINSTANCE hDllHandle, IN DWORD nReason,
                        IN LPVOID Reserved)
 {
+    jl_task_t *ct;
     switch (nReason) {
     case DLL_PROCESS_ATTACH:
         jl_pgcstack_key = TlsAlloc();
         assert(jl_pgcstack_key != TLS_OUT_OF_INDEXES);
         jl_safe_restore_key = TlsAlloc();
         assert(jl_safe_restore_key != TLS_OUT_OF_INDEXES);
-        // Fall through
-    case DLL_THREAD_ATTACH:
-        break;
-    case DLL_THREAD_DETACH:
         break;
     case DLL_PROCESS_DETACH:
         TlsFree(jl_pgcstack_key);
         TlsFree(jl_safe_restore_key);
         break;
+    case DLL_THREAD_ATTACH:
+        // will call jl_adopt_thread lazily on-demand
+        break;
+    case DLL_THREAD_DETACH:
+        ct = jl_get_current_task();
+        if (ct != NULL)
+            jl_delete_thread((void*)ct->ptls);
+        break;
     }
     return 1; // success
 }
@@ -241,7 +254,7 @@ static jl_gcframe_t **jl_get_pgcstack_init(void)
     // are used. Since the address of TLS variables should be constant,
     // changing the getter address can result in weird crashes.
 
-    // This is clearly not thread safe but should be fine since we
+    // This is clearly not thread-safe but should be fine since we
     // make sure the tls states callback is finalized before adding
     // multiple threads
 #  if JL_USE_IFUNC
@@ -278,32 +291,57 @@ JL_DLLEXPORT jl_gcframe_t **jl_get_pgcstack(void) JL_GLOBALLY_ROOTED
 
 void jl_pgcstack_getkey(jl_get_pgcstack_func **f, jl_pgcstack_key_t *k)
 {
+#ifndef __clang_gcanalyzer__
     if (jl_get_pgcstack_cb == jl_get_pgcstack_init)
         jl_get_pgcstack_init();
+#endif
     // for codegen
     *f = jl_get_pgcstack_cb;
     *k = jl_pgcstack_key;
 }
 #endif
 
-jl_ptls_t *jl_all_tls_states JL_GLOBALLY_ROOTED;
-JL_DLLEXPORT _Atomic(uint8_t) jl_measure_compile_time_enabled = 0;
-JL_DLLEXPORT _Atomic(uint64_t) jl_cumulative_compile_time = 0;
-JL_DLLEXPORT _Atomic(uint64_t) jl_cumulative_recompile_time = 0;
+static uv_mutex_t tls_lock; // controls write-access to these variables:
+_Atomic(jl_ptls_t*) jl_all_tls_states JL_GLOBALLY_ROOTED;
+int jl_all_tls_states_size;
+static uv_cond_t cond;
 
 // return calling thread's ID
-// Also update the suspended_threads list in signals-mach when changing the
-// type of the thread id.
 JL_DLLEXPORT int16_t jl_threadid(void)
 {
     return jl_atomic_load_relaxed(&jl_current_task->tid);
 }
 
+JL_DLLEXPORT int8_t jl_threadpoolid(int16_t tid) JL_NOTSAFEPOINT
+{
+    int nthreads = jl_atomic_load_acquire(&jl_n_threads);
+    if (tid < 0 || tid >= nthreads)
+        jl_error("invalid tid");
+    int n = 0;
+    for (int i = 0; i < jl_n_threadpools; i++) {
+        n += jl_n_threads_per_pool[i];
+        if (tid < n)
+            return (int8_t)i;
+    }
+    return 0; // everything else uses threadpool 0 (though does not become part of any threadpool)
+}
+
 jl_ptls_t jl_init_threadtls(int16_t tid)
 {
+#ifndef _OS_WINDOWS_
+    if (pthread_getspecific(jl_task_exit_key))
+        abort();
+#endif
+    if (jl_get_pgcstack() != NULL)
+        abort();
     jl_ptls_t ptls = (jl_ptls_t)calloc(1, sizeof(jl_tls_states_t));
+#ifndef _OS_WINDOWS_
+    pthread_setspecific(jl_task_exit_key, (void*)ptls);
+#endif
     ptls->system_id = (jl_thread_t)(uintptr_t)uv_thread_self();
     ptls->rngseed = jl_rand();
+    if (tid == 0)
+        ptls->disable_gc = 1;
 #ifdef _OS_WINDOWS_
     if (tid == 0) {
         if (!DuplicateHandle(GetCurrentProcess(), GetCurrentThread(),
@@ -314,7 +352,6 @@ jl_ptls_t jl_init_threadtls(int16_t tid)
         }
     }
 #endif
-    ptls->tid = tid;
     jl_atomic_store_relaxed(&ptls->gc_state, 0); // GC unsafe
     // Conditionally initialize the safepoint address. See comment in
     // `safepoint.c`
@@ -332,11 +369,83 @@ jl_ptls_t jl_init_threadtls(int16_t tid)
     small_arraylist_new(&ptls->locks, 0);
     jl_init_thread_heap(ptls);
 
-    jl_all_tls_states[tid] = ptls;
+    uv_mutex_init(&ptls->sleep_lock);
+    uv_cond_init(&ptls->wake_signal);
+
+    uv_mutex_lock(&tls_lock);
+    jl_ptls_t *allstates = jl_atomic_load_relaxed(&jl_all_tls_states);
+    if (tid == -1)
+        tid = jl_atomic_load_relaxed(&jl_n_threads);
+    ptls->tid = tid;
+    if (jl_all_tls_states_size <= tid) {
+        int i, newsize = jl_all_tls_states_size + tid + 2;
+        jl_ptls_t *newpptls = (jl_ptls_t*)calloc(newsize, sizeof(jl_ptls_t));
+        for (i = 0; i < jl_all_tls_states_size; i++) {
+            newpptls[i] = allstates[i];
+        }
+        jl_atomic_store_release(&jl_all_tls_states, newpptls);
+        jl_all_tls_states_size = newsize;
+        jl_gc_add_quiescent(ptls, (void**)allstates, free);
+        allstates = newpptls;
+    }
+    allstates[tid] = ptls;
+    if (jl_atomic_load_relaxed(&jl_n_threads) < tid + 1)
+        jl_atomic_store_release(&jl_n_threads, tid + 1);
+    jl_fence();
+    uv_mutex_unlock(&tls_lock);
 
     return ptls;
 }
 
+JL_DLLEXPORT jl_gcframe_t **jl_adopt_thread(void)
+{
+    // initialize this thread (assign tid, create heap, set up root task)
+    jl_ptls_t ptls = jl_init_threadtls(-1);
+    void *stack_lo, *stack_hi;
+    jl_init_stack_limits(0, &stack_lo, &stack_hi);
+
+    (void)jl_gc_unsafe_enter(ptls);
+    // warning: this changes `jl_current_task`, so be careful not to call that from this function
+    jl_task_t *ct = jl_init_root_task(ptls, stack_lo, stack_hi);
+    JL_GC_PROMISE_ROOTED(ct);
+    uv_random(NULL, NULL, &ct->rngState, sizeof(ct->rngState), 0, NULL);
+    return &ct->gcstack;
+}
+
+static void jl_delete_thread(void *value)
+{
+    jl_ptls_t ptls = (jl_ptls_t)value;
+    // Acquire the profile write lock, to ensure we are not racing with the `kill`
+    // call in the profile code which will also try to look at these variables.
+    // We have no control over when the user calls pthread_join, so we must do
+    // this here by blocking. This also synchronizes our read of `current_task`
+    // (which is the flag we currently use to check the liveness state of a thread).
+#ifdef _OS_WINDOWS_
+    jl_lock_profile_wr();
+#elif defined(JL_DISABLE_LIBUNWIND)
+    // nothing
+#elif defined(__APPLE__)
+    jl_lock_profile_wr();
+#else
+    pthread_mutex_lock(&in_signal_lock);
+#endif
+#ifndef _OS_WINDOWS_
+    pthread_setspecific(jl_task_exit_key, NULL);
+#endif
+    jl_atomic_store_relaxed(&ptls->current_task, NULL); // dead
+    jl_atomic_store_relaxed(&ptls->sleep_check_state, 2); // dead, interpreted as sleeping and unwakeable
+#ifdef _OS_WINDOWS_
+    jl_unlock_profile_wr();
+#elif defined(JL_DISABLE_LIBUNWIND)
+    // nothing
+#elif defined(__APPLE__)
+    jl_unlock_profile_wr();
+#else
+    pthread_mutex_unlock(&in_signal_lock);
+#endif
+    (void)jl_gc_safe_enter(ptls);
+}
+
 JL_DLLEXPORT jl_mutex_t jl_codegen_lock;
 jl_mutex_t typecache_lock;
 
@@ -448,41 +557,73 @@ void jl_init_threading(void)
 {
     char *cp;
 
+    uv_mutex_init(&tls_lock);
+    uv_cond_init(&cond);
 #ifdef JL_ELF_TLS_VARIANT
     jl_check_tls();
 #endif
 
-    // how many threads available, usable
-    jl_n_threads = JULIA_NUM_THREADS;
-    if (jl_options.nthreads < 0) { // --threads=auto
-        jl_n_threads = jl_cpu_threads();
+    // Determine how many threads and pools are requested. This may have been
+    // specified on the command line (and so are in `jl_options`) or by the
+    // environment variable. Set the globals `jl_n_threadpools`, `jl_n_threads`
+    // and `jl_n_threads_per_pool`.
+    jl_n_threadpools = 2;
+    int16_t nthreads = JULIA_NUM_THREADS;
+    int16_t nthreadsi = 0;
+    char *endptr, *endptri;
+
+    if (jl_options.nthreads != 0) { // --threads specified
+        nthreads = jl_options.nthreads_per_pool[0];
+        if (nthreads < 0)
+            nthreads = jl_effective_threads();
+        if (jl_options.nthreadpools == 2)
+            nthreadsi = jl_options.nthreads_per_pool[1];
     }
-    else if (jl_options.nthreads > 0) { // --threads=N
-        jl_n_threads = jl_options.nthreads;
+    else if ((cp = getenv(NUM_THREADS_NAME))) { // ENV[NUM_THREADS_NAME] specified
+        if (!strncmp(cp, "auto", 4)) {
+            nthreads = jl_effective_threads();
+            cp += 4;
+        }
+        else {
+            errno = 0;
+            nthreads = strtol(cp, &endptr, 10);
+            if (errno != 0 || endptr == cp || nthreads <= 0)
+                nthreads = 1;
+            cp = endptr;
+        }
+        if (*cp == ',') {
+            cp++;
+            if (!strncmp(cp, "auto", 4))
+                nthreadsi = 1;
+            else {
+                errno = 0;
+                nthreadsi = strtol(cp, &endptri, 10);
+                if (errno != 0 || endptri == cp || nthreadsi < 0)
+                    nthreadsi = 0;
+            }
+        }
     }
-    else if ((cp = getenv(NUM_THREADS_NAME))) {
-        if (strcmp(cp, "auto"))
-            jl_n_threads = (uint64_t)strtol(cp, NULL, 10); // ENV[NUM_THREADS_NAME] == "N"
-        else
-            jl_n_threads = jl_cpu_threads(); // ENV[NUM_THREADS_NAME] == "auto"
-    }
-    if (jl_n_threads <= 0)
-        jl_n_threads = 1;
-#ifndef __clang_gcanalyzer__
-    jl_all_tls_states = (jl_ptls_t*)calloc(jl_n_threads, sizeof(void*));
-#endif
+
+    jl_all_tls_states_size = nthreads + nthreadsi;
+    jl_n_threads_per_pool = (int*)malloc_s(2 * sizeof(int));
+    jl_n_threads_per_pool[0] = nthreadsi;
+    jl_n_threads_per_pool[1] = nthreads;
+
+    jl_atomic_store_release(&jl_all_tls_states, (jl_ptls_t*)calloc(jl_all_tls_states_size, sizeof(jl_ptls_t)));
+    jl_atomic_store_release(&jl_n_threads, jl_all_tls_states_size);
 }
 
 static uv_barrier_t thread_init_done;
 
 void jl_start_threads(void)
 {
+    int nthreads = jl_atomic_load_relaxed(&jl_n_threads);
     int cpumasksize = uv_cpumask_size();
     char *cp;
     int i, exclusive;
     uv_thread_t uvtid;
-    if (cpumasksize < jl_n_threads) // also handles error case
-        cpumasksize = jl_n_threads;
+    if (cpumasksize < nthreads) // also handles error case
+        cpumasksize = nthreads;
     char *mask = (char*)alloca(cpumasksize);
 
     // do we have exclusive use of the machine? default is no
@@ -495,7 +636,7 @@ void jl_start_threads(void)
     // according to a 'compact' policy
     // non-exclusive: no affinity settings; let the kernel move threads about
     if (exclusive) {
-        if (jl_n_threads > jl_cpu_threads()) {
+        if (nthreads > jl_cpu_threads()) {
             jl_printf(JL_STDERR, "ERROR: Too many threads requested for %s option.\n", MACHINE_EXCLUSIVE_NAME);
             exit(1);
         }
@@ -506,14 +647,11 @@ void jl_start_threads(void)
         mask[0] = 0;
     }
 
-    // The analyzer doesn't know jl_n_threads doesn't change, help it
-    size_t nthreads = jl_n_threads;
-
     // create threads
     uv_barrier_init(&thread_init_done, nthreads);
 
     for (i = 1; i < nthreads; ++i) {
-        jl_threadarg_t *t = (jl_threadarg_t*)malloc_s(sizeof(jl_threadarg_t)); // ownership will be passed to the thread
+        jl_threadarg_t *t = (jl_threadarg_t *)malloc_s(sizeof(jl_threadarg_t)); // ownership will be passed to the thread
         t->tid = i;
         t->barrier = &thread_init_done;
         uv_thread_create(&uvtid, jl_threadfun, t);
@@ -553,6 +691,113 @@ JL_DLLEXPORT void jl_exit_threaded_region(void)
     }
 }
 
+void _jl_mutex_wait(jl_task_t *self, jl_mutex_t *lock, int safepoint)
+{
+    jl_task_t *owner = jl_atomic_load_relaxed(&lock->owner);
+    if (owner == self) {
+        lock->count++;
+        return;
+    }
+    while (1) {
+        if (owner == NULL && jl_atomic_cmpswap(&lock->owner, &owner, self)) {
+            lock->count = 1;
+            return;
+        }
+        if (safepoint) {
+            jl_gc_safepoint_(self->ptls);
+        }
+        if (jl_running_under_rr(0)) {
+            // when running under `rr`, use system mutexes rather than spin locking
+            uv_mutex_lock(&tls_lock);
+            if (jl_atomic_load_relaxed(&lock->owner))
+                uv_cond_wait(&cond, &tls_lock);
+            uv_mutex_unlock(&tls_lock);
+        }
+        jl_cpu_pause();
+        owner = jl_atomic_load_relaxed(&lock->owner);
+    }
+}
+
+static void jl_lock_frame_push(jl_task_t *self, jl_mutex_t *lock)
+{
+    jl_ptls_t ptls = self->ptls;
+    small_arraylist_t *locks = &ptls->locks;
+    uint32_t len = locks->len;
+    if (__unlikely(len >= locks->max)) {
+        small_arraylist_grow(locks, 1);
+    }
+    else {
+        locks->len = len + 1;
+    }
+    locks->items[len] = (void*)lock;
+}
+
+static void jl_lock_frame_pop(jl_task_t *self)
+{
+    jl_ptls_t ptls = self->ptls;
+    assert(ptls->locks.len > 0);
+    ptls->locks.len--;
+}
+
+void _jl_mutex_lock(jl_task_t *self, jl_mutex_t *lock)
+{
+    JL_SIGATOMIC_BEGIN_self();
+    _jl_mutex_wait(self, lock, 1);
+    jl_lock_frame_push(self, lock);
+}
+
+int _jl_mutex_trylock_nogc(jl_task_t *self, jl_mutex_t *lock)
+{
+    jl_task_t *owner = jl_atomic_load_acquire(&lock->owner);
+    if (owner == self) {
+        lock->count++;
+        return 1;
+    }
+    if (owner == NULL && jl_atomic_cmpswap(&lock->owner, &owner, self)) {
+        lock->count = 1;
+        return 1;
+    }
+    return 0;
+}
+
+int _jl_mutex_trylock(jl_task_t *self, jl_mutex_t *lock)
+{
+    int got = _jl_mutex_trylock_nogc(self, lock);
+    if (got) {
+        JL_SIGATOMIC_BEGIN_self();
+        jl_lock_frame_push(self, lock);
+    }
+    return got;
+}
+
+void _jl_mutex_unlock_nogc(jl_mutex_t *lock)
+{
+#ifndef __clang_gcanalyzer__
+    assert(jl_atomic_load_relaxed(&lock->owner) == jl_current_task &&
+           "Unlocking a lock in a different thread.");
+    if (--lock->count == 0) {
+        jl_atomic_store_release(&lock->owner, (jl_task_t*)NULL);
+        jl_cpu_wake();
+        if (jl_running_under_rr(0)) {
+            // when running under `rr`, use system mutexes rather than spin locking
+            uv_mutex_lock(&tls_lock);
+            uv_cond_broadcast(&cond);
+            uv_mutex_unlock(&tls_lock);
+        }
+    }
+#endif
+}
+
+void _jl_mutex_unlock(jl_task_t *self, jl_mutex_t *lock)
+{
+    _jl_mutex_unlock_nogc(lock);
+    jl_lock_frame_pop(self);
+    JL_SIGATOMIC_END_self();
+    if (jl_atomic_load_relaxed(&jl_gc_have_pending_finalizers)) {
+        jl_gc_run_pending_finalizers(self); // may GC
+    }
+}
+
 
 // Make gc alignment available for threading
 // see threads.jl alignment
