@@ -288,7 +288,7 @@ end
 end
 @testset "issue #21510" begin
     f21510(; @nospecialize a = 2) = a
-    @test f21510(a=:b) == :b
+    @test f21510(a=:b) === :b
     @test f21510() == 2
 end
 @testset "issue #34516" begin
@@ -297,7 +297,7 @@ end
     @test_throws UndefKeywordError f34516()
     @test_throws UndefKeywordError f34516(1)
     g34516(@nospecialize(x); k=0) = 0
-    @test first(methods(Core.kwfunc(g34516))).nospecialize != 0
+    @test only(methods(Core.kwcall, (Any, typeof(g34516), Vararg))).nospecialize != 0
 end
 @testset "issue #21518" begin
     a = 0
@@ -387,3 +387,10 @@ f41416(a...="a"; b=true) = (b, a)
 @test f41416(;b=false)   === (false, ("a",))
 @test f41416(33)         === (true, (33,))
 @test f41416(3; b=false) === (false, (3,))
+
+Core.kwcall(i::Int) = "hi $i"
+let m = first(methods(Core.kwcall, (Any,typeof(kwf1),Vararg)))
+    @test m.name === :kwf1
+    @test Core.kwcall(1) == "hi 1"
+    @test which(Core.kwcall, (Int,)).name === :kwcall
+end
